%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "asterisk/argdesc.h"
#include "asterisk/logger.h"

#define YYPARSE_PARAM parseio
#define YYLEX_PARAM ((struct parse_arginfo *)parseio)->scanner
#define YYERROR_VERBOSE 1
/* this stuff is included so I can declare the proto for yyerror */ 
#if ! defined (YYLTYPE) && ! defined (YYLTYPE_IS_DECLARED)

typedef struct YYLTYPE
{
  int first_line;
  int first_column;
  int last_line;
  int last_column;
} YYLTYPE;

# define yyltype YYLTYPE /* obsolescent; will be withdrawn */
# define YYLTYPE_IS_DECLARED 1
# define YYLTYPE_IS_TRIVIAL 1
#endif
#if ! defined (YYSTYPE) && ! defined (YYSTYPE_IS_DECLARED)
typedef union YYSTYPE {
	char *str;
	struct argapp *argapp;
	struct argdesc *argdesc;
	struct appsetvar *appsetvar;
	struct argchoice *argchoice;
} YYSTYPE;
# define yystype YYSTYPE /* obsolescent; will be withdrawn */
# define YYSTYPE_IS_DECLARED 1
# define YYSTYPE_IS_TRIVIAL 1
#endif

 static int errcnt = 0;
 
void yyerror(YYLTYPE *locp,void *nothing, char const *s);
int argdesc_yylex (YYSTYPE * yylval_param, YYLTYPE * yylloc_param , void * yyscanner);
void linkarg(struct argdesc *head, struct argdesc *tail);
static void argdesc_destroy_argdesc(struct argdesc *desc);
static void argdesc_destroy_appsetvar(struct appsetvar *var);
static void argdesc_destroy_argchoice(struct argchoice *choice);
static void argdesc_destroy_argapp(struct argapp *app);
 
%}


%union {
	char *str;
	struct argapp *argapp;
	struct argdesc *argdesc;
	struct appsetvar *appsetvar;
	struct argchoice *argchoice;
}


%token SEMI COMMA BAR EQ LP RP LB RB LAB RAB COLON LC RC ELLIPSES
%token KW_ARGS KW_OPTIONS KW_CHANVARS intspec fltspec strspec

%token <str> word str1

%type <argapp> file appentries appentry
%type <argdesc> arglist argdeclist argdec argspec
%type <appsetvar> varlist varspec varspeclist
%type <argchoice> choicelist vallist optionslist optionspeclist optionspec

/* OPTIONS */
%locations
%pure-parser
%name-prefix="argdesc_yy"
/* the following option does two things: 
    it adds the locp arg to the yyerror
    and it adds the NULL to the yyerrr arg list, and calls yyerror with NULL for that arg.
    You can't get the locp arg without the NULL arg, don't ask me why. */
%parse-param {void *NULL}

%error-verbose

%%

file : appentries   {((struct parse_arginfo *)parseio)->apps = $$; ((struct parse_arginfo *)parseio)->syntax_error_count = errcnt;}
	;

appentries: appentry   {$$=$1;}
	| appentries appentry   {$2->next = $1; $$ = $2; }
	;

appentry: word COLON str1 SEMI arglist optionslist varlist  {
	$$=(struct argapp *)calloc(sizeof(struct argapp),1);
	$$->name = $1;
	$$->desc = $3;
	$$->args = $5;
	$$->opts = $6;
	$$->setvars = $7;
		}
	;

arglist : /* nothing */   {$$=0;}
		| KW_ARGS EQ argdeclist SEMI   {$$=$3;}
		;

argdeclist : argdec   {$$=$1;}
		| argdeclist BAR argdec   {$$=$1; linkarg($$,$3);}
		;

argdec : argspec    {$$=$1; $$->type=ARGD_REQUIRED;}
		| LB argspec RB   {$$=$2; $$->type=ARGD_OPTIONAL;}
		| LB COLON argspec RB   {$$=$3; $$->type=ARGD_OPTIONAL_PREV;}
		| LC argspec RC   {$$=$2; $$->type=ARGD_OPTIONAL_PLACEHOLD;}
		;
argspec : word    {$$=(struct argdesc*)calloc(sizeof(struct argdesc),1); $$->name=$1;$$->dtype=ARGD_STRING;}
		| word intspec   {$$=(struct argdesc*)calloc(sizeof(struct argdesc),1); $$->name=$1; $$->dtype=ARGD_INT;}
		| word strspec   {$$=(struct argdesc*)calloc(sizeof(struct argdesc),1); $$->name=$1;$$->dtype=ARGD_STRING;}
		| word fltspec   {$$=(struct argdesc*)calloc(sizeof(struct argdesc),1); $$->name=$1;$$->dtype=ARGD_FLOAT;}
		| ELLIPSES      {$$=(struct argdesc*)calloc(sizeof(struct argdesc),1); $$->name=0;$$->dtype=ARGD_VARARG;}
		| KW_OPTIONS   {$$=(struct argdesc*)calloc(sizeof(struct argdesc),1); $$->name=0;$$->dtype=ARGD_OPTIONSET;}
		| LAB word COLON choicelist RAB   {$$=(struct argdesc*)calloc(sizeof(struct argdesc),1); $$->name=$2;$$->dtype=ARGD_ENUM;$$->choices = $4;}
		;

choicelist : word   {$$=(struct argchoice*)calloc(sizeof(struct argchoice),1); $$->name=$1;}
		| choicelist COMMA word   {$$=(struct argchoice*)calloc(sizeof(struct argchoice),1); $$->name=$3; $$->next = $1;}
		;

optionslist : KW_OPTIONS EQ optionspeclist SEMI   {$$=$3;}
		| /* nothing */   {$$=0;}
		;

optionspeclist : optionspec   {$$=$1;}
				| optionspeclist COMMA optionspec   {$$=$3; $3->next = $1; }
				;

optionspec : word   {$$ = (struct argchoice *)calloc(sizeof(struct argchoice),1); $$->name=$1; $$->opttype=ARGD_OPT_NOARG;}
		| word LP RP   {$$ = (struct argchoice *)calloc(sizeof(struct argchoice),1); $$->name=$1; $$->opttype=ARGD_OPT_WITHARG;}
		| word LB LP RP RB   {$$ = (struct argchoice *)calloc(sizeof(struct argchoice),1); $$->name=$1; $$->opttype=ARGD_OPT_WITHOPTARG;}
		;


varlist : /* nothing */   {$$=0;}
		| KW_CHANVARS EQ varspeclist SEMI   {$$=$3;}
		;

varspeclist : varspec   {$$=$1;}
			| varspeclist COMMA varspec   {$$ = $3; $$->next = $1;}
			;

varspec : word   {$$=(struct appsetvar*)calloc(sizeof(struct appsetvar),1); $$->name=$1;}
		| word COLON vallist    {$$=(struct appsetvar*)calloc(sizeof(struct appsetvar),1); $$->name=$1; $$->vals=$3;}
		;

vallist : word   {$$=(struct argchoice*)calloc(sizeof(struct argchoice),1); $$->name=$1; }
		| vallist BAR word {$$=(struct argchoice*)calloc(sizeof(struct argchoice),1); $$->name=$3; $$->next = $1; }
		;


%%

void yyerror(YYLTYPE *locp, void *nothing,  char const *s)
{
	
	if (locp->first_line == locp->last_line) {
		ast_log(LOG_ERROR, "==== Line %d, Cols: %d-%d: Error: %s\n", locp->first_line, locp->first_column, locp->last_column, s);
	} else {
		ast_log(LOG_ERROR, "==== Line %d Col %d  to Line %d Col %d: Error: %s\n", locp->first_line, locp->first_column, locp->last_line, locp->last_column, s);
	}
	errcnt++;
}

void linkarg(struct argdesc *head, struct argdesc *tail)
{
	if (!head->next) {
		head->next = tail;
		head->last_args = tail;
	} else {
		head->last_args->next = tail;
		head->last_args = tail;
	}
}

static void argdesc_destroy_argdesc(struct argdesc *desc)
{
	struct argchoice *c, *cn;
	if (desc->name)
		free(desc->name);
	for(c=desc->choices; c;) {
		cn = c->next;
		argdesc_destroy_argchoice(c);
		c = cn;
	}
	desc->name = 0;
	desc->choices = 0;
	desc->next = 0;
	free(desc);
}

static void argdesc_destroy_appsetvar(struct appsetvar *var)
{
	struct argchoice *c, *cn;
	if (var->name)
		free(var->name);
	for (c=var->vals; c;) {
		cn = c->next;
		argdesc_destroy_argchoice(c);
		c = cn;
	}
	var->name = 0;
	var->vals = 0;
	var->next = 0;
	free(var);
}

static void argdesc_destroy_argchoice(struct argchoice *choice)
{
	if( choice->name )
		free( choice->name);
	choice->name = 0;
	choice->next = 0;
	free(choice);
}


static void argdesc_destroy_argapp(struct argapp *app)
{
	struct argdesc *d,*dn;
	struct appsetvar *v, *vn;
	struct argchoice *c, *cn;
	
	for (d=app->args; d; ) {
		dn = d->next;
		argdesc_destroy_argdesc(d);
		d = dn;
	}
	for (v=app->setvars; v;) {
		vn = v->next;
		argdesc_destroy_appsetvar(v);
		v = vn;
	}
	for(c=app->opts; c;) {
		cn = c->next;
		argdesc_destroy_argchoice(c);
		c = cn;
	}
	if (app->name)
		free(app->name);
	if (app->desc)
		free(app->desc);
	app->name = 0;
	app->desc = 0;
	app->next = 0;
	app->args = 0;
	app->setvars = 0;
	app->opts = 0;
	free(app);
}

void argdesc_destroy(struct argapp *apps)
{
	struct argapp *p,*pn;
	
	for (p=apps; p; ) {
		pn = p->next;
		argdesc_destroy_argapp(p);
		p = pn;
	}
}
